from . import GraderReporter
from gradema.section import Section
from gradema.test import Test, FractionalTestResult, PercentTestResult


def point_sections(max_points: int, sections: list[Section]) -> list[int]:
    point_values_sum = 0
    point_values: list[int] = [0 for _ in range(len(sections))]
    weight_total = 0.0
    evenly_weighted_count = 0
    for i, section in enumerate(sections):
        if section.is_pointed or section.is_ungraded:
            assert section.points is not None
            point_values[i] = section.points
            point_values_sum += section.points
        elif section.is_weighted:
            assert section.weight is not None
            weight_total += section.weight
        else:
            assert section.is_evenly_weighted
            evenly_weighted_count += 1
    if evenly_weighted_count > 0:
        evenly_weighted_weight = (1 - weight_total) / evenly_weighted_count
        if evenly_weighted_weight <= 0:
            raise ValueError(
                f"weight_total: {weight_total} which is invalid when you have evenly weighted section. "
                f"Make sure it is less than 1 so the weight can be distributed evenly among those sections."
            )
        weight_divisor = 1.0
    else:
        evenly_weighted_weight = 0.0
        weight_divisor = weight_total

    remaining_points = max_points - point_values_sum
    if remaining_points < 0:
        raise ValueError(f"The sum of all the pointed sections are higher than max points. max_points: {max_points} sum: {sum(point_values)}")

    points_assigned = 0  # the sum of the points assigned to weighted or evenly weighted sections
    running_weight_sum = 0.0
    for i, section in enumerate(sections):
        if section.is_weighted or section.is_evenly_weighted:
            if section.is_weighted:
                assert section.weight is not None
                weight = section.weight / weight_divisor
            else:
                weight = evenly_weighted_weight
            running_weight_sum += weight
            target_points_assigned = remaining_points * running_weight_sum
            points = int(target_points_assigned - points_assigned)
            point_values[i] = points
            points_assigned += points

    assert points_assigned == remaining_points
    return point_values


def grade_test(reporter: GraderReporter, max_points: int, test: Test) -> int:
    reporter.report_start(max_points)
    result = test.run(reporter.test_reporter)
    if isinstance(result, FractionalTestResult):
        points = max_points * result.success_count // result.total_count
    else:
        assert isinstance(result, PercentTestResult)
        points = int(result.percent * max_points)

    reporter.report_test_result(result, points, max_points)
    return points


def grade_section(reporter: GraderReporter, max_points: int, section: Section) -> int:
    if isinstance(section.node, Test):
        return grade_test(reporter, max_points, section.node)
    assert isinstance(section.node, list)
    sections: list[Section] = section.node
    assert len(sections) > 0
    assert isinstance(sections[0], Section)
    point_values = point_sections(max_points, sections)

    reporter.report_start(max_points)
    total_points_earned = 0
    for i, section in enumerate(sections):
        points = point_values[i]
        subsection_reporter = reporter.subsection(section)
        points_received = grade_section(subsection_reporter, points, section)
        total_points_earned += points_received
    return total_points_earned



import importlib.util
from pathlib import Path
from typing import Sequence, Optional

import editorconfig  # type: ignore

from . import Test
from ._command import CommandTest
from ._stdio import StdioTest
from .argument import Argument


def create_python_test(module_name: str) -> Test:
    """
    Creates a python test that runs the given module and is successful depending on the successful exit code of running the given module.

    Please note that this kind of test is susceptible to "early exit attacks".
    Details here: https://gradema.readthedocs.io/en/latest/develop/assignment/security.html#early-exit-attacks

    :param module_name: The name of the module to run
    :return: A test whose success is based on the exit code of running the specified module
    """
    if importlib.util.find_spec(module_name) is None:
        raise ValueError(f"Module {module_name} not found")

    return CommandTest(["python", "-m", module_name], ["pudb", "-m", module_name])


def create_python_pytest(file_argument: str) -> Test:
    """
    Creates a pytest test.

    Currently, does not have a debugging option. See source code of this function for detailed comment.

    :param file_argument: The argument to pass to pytest that is the file and optionally has a ``::test_function_name`` tacked on at the end
    :return: A pytest test
    """
    # TODO add a debug command to pytest
    """
    So... I (Lavender Shannon) spent a couple of hours trying to figure out how to use pudb and pytest together.

    If you do something like ``pudb -m pytest example/binaryconvert/autograder/test/unit/test_convert_1.py::test_convert_1`` it drops you into pytest's main.
    You can eventually get dropped into debugging the function you actually care about, but it's not intuitive.

    You can try something like ``pytest --pdb --pdbcls=pudb.debugger:Debugger example/binaryconvert/autograder/test/unit/test_convert_5.py::test_convert_5``,
    and that works reasonably well only if you want to view the stacktrace within pudb. You don't actually get to debug the function.

    The only thing that works reasonably well is using https://github.com/wronglink/pytest-pudb and set pudb.set_trace() calls at the beginning of your test function.
    Problem with this is that it seems glitchy after you are done debugging the function you care about. Like it's running it multiple times? I'm not sure why...
    """

    # return CommandTest(["pytest", "-q", "--no-summary", "--no-header", file_argument])
    return CommandTest(["pytest", "-q", "--no-header", file_argument])


def create_python_stdio_test(module_name: str, test_identifier: str, args: Sequence[Argument], input_file: Optional[Path], goal_file: Path, stdout_as_output: bool) -> Test:
    command: list[Argument] = ["python", "-m", module_name]
    command.extend(args)
    return StdioTest(command, test_identifier, input_file, goal_file, stdout_as_output)


def create_python_traditional_stdio_test(module_name: str, test_identifier: str, input_file: Path, goal_file: Path) -> Test:
    """
    Creates a traditional stdio test

    :param module_name:
    :param test_identifier:
    :param input_file:
    :param goal_file:
    :return:
    """
    return create_python_stdio_test(module_name, test_identifier, [], input_file, goal_file, stdout_as_output=True)


def create_python_traditional_arg_test(module_name: str, test_identifier: str, args: Sequence[Argument], goal_file: Path) -> Test:
    return create_python_stdio_test(module_name, test_identifier, args, None, goal_file, stdout_as_output=False)


def create_python_format_check_from_path(path: str) -> Test:
    line_length = editorconfig.get_properties(str((Path(path) / "some_random_file.py").absolute())).get("max_line_length")
    command = ["black", "--check"]
    if line_length is not None:
        command.append(f"--line-length={line_length}")
    command.append(path)
    return CommandTest(command)


def create_python_format_check() -> Test:
    return create_python_format_check_from_path(".")


def create_python_type_check() -> Test:
    return CommandTest(["mypy", "--strict", "--disallow-any-explicit", "."])


def create_mypy_command(arguments: Sequence[str]) -> Test:
    return CommandTest(["mypy", *arguments])



import copy
import dataclasses
import difflib
import io
import shlex
import shutil
import subprocess
import traceback
import webbrowser
from pathlib import Path
from typing import TextIO, Optional, Sequence, Mapping

from rich import markup
from rich.console import Console
from rich.prompt import Prompt

from gradema.grader import GraderReporter
from gradema.grader.console._util import indentation, arrow, header, indent_text
from gradema.section import Section
from gradema.test import (
    TestReporter,
    CommandData,
    Test,
    TestResult,
    FractionalTestResult,
    PercentTestResult,
    StdioCommandData,
)
from gradema.test.argument import Argument, ResolvedArgumentInfo, OutputFile

TEST_DIRECTORY = Path("build/test")


def has_vim_installed() -> bool:
    return shutil.which("vim") is not None


def vim_diff(goal_file: Path, output_file: Path, html_output: Path) -> None:
    """
    Creates a diff using vim.
    Note that calling this function takes a while to complete.

    :param goal_file:
    :param output_file:
    :param html_output:
    """
    subprocess.run(
        [
            "vim",
            "-d",
            str(goal_file),
            str(output_file),
            "-c",
            "highlight DiffAdd ctermbg=Grey ctermfg=White",
            "-c",
            "highlight DiffDelete ctermbg=Grey ctermfg=Black",
            "-c",
            "highlight DiffChange ctermbg=Grey ctermfg=Black",
            "-c",
            "highlight DiffText ctermbg=DarkGrey ctermfg=White",
            "-c",
            "TOhtml",
            "-c",
            f"w! {str(html_output)}",
            "-c",
            "qa!",
        ],
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )


def html_diff(goal_path: Path, actual_output_path: Path, html_output_path: Path) -> None:
    # TODO remember that file.readlines() LOADS THE ENTIRE FILE INTO MEMORY
    #   We might not care about that right now, but this could be a problem later with huge files
    with goal_path.open("r") as left, actual_output_path.open("r") as right:
        diff = difflib.HtmlDiff().make_file(
            left.readlines(),
            right.readlines(),
            f"Goal: {goal_path}",
            f"Your output: {actual_output_path}",
        )
        with html_output_path.open("w") as f_out:
            f_out.write(diff)


@dataclasses.dataclass
class ConsoleTestReporter(TestReporter):
    reporter: "ConsoleGraderReporter"

    def log(self, message: str) -> None:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(indent + markup.escape(message))

    def log_unexpected_exception(self, exception: BaseException) -> None:
        indent = indentation(self.reporter.depth)
        buffer = io.StringIO()
        traceback.print_exception(exception, file=buffer)
        indented_text = indent_text(self.reporter.depth, buffer.getvalue())
        # make end="" because the exception text should have a trailing newline
        self.reporter.console.print(f"{indent}Unexpected exception:\n{indented_text}", highlight=False, end="")

    def __get_output_file(self, test_identifier: str) -> Path:
        directory = TEST_DIRECTORY / "stdio/outputs"
        directory.mkdir(parents=True, exist_ok=True)
        return directory / f"{test_identifier}.txt"

    def resolve_command(self, command: Sequence[Argument], test_identifier: str) -> tuple[Sequence[str], Mapping[tuple[int, Argument], ResolvedArgumentInfo]]:
        resolved_command = []
        argument_info_dictionary: dict[tuple[int, Argument], ResolvedArgumentInfo] = dict()
        # NOTE: We have to specify the type of argument_info_dictionary because the key is not covariant: https://github.com/python/typing/pull/273
        for i, argument in enumerate(command):
            if isinstance(argument, OutputFile):
                output_file = self.__get_output_file(test_identifier)
                resolved_command.append(str(output_file))
                argument_info_dictionary[(i, argument)] = output_file
            else:  # argument should be a str. If you get errors here, someone messed with the definition of Argument and you need to implement more cases
                resolved_command.append(argument)

        return resolved_command, argument_info_dictionary

    def report_command(self, command: Sequence[str]) -> CommandData:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(f"{indent}Running command: [magenta]{markup.escape(shlex.join(command))}[/magenta]")
        return CommandData(self.reporter.stdout, self.reporter.stderr)

    def report_stdio_command(self, command: Sequence[str], test_identifier: str, input_file: Optional[Path], stdout_as_output: bool) -> StdioCommandData:
        indent = indentation(self.reporter.depth)
        indent_plus = indentation(self.reporter.depth + 1)
        output_file = self.__get_output_file(test_identifier)
        # NOTE: On Windows shlex.join will almost always result in quoting the entire argument.
        #   Because of this, we need to turn off automatic highlighting, or it won't be magenta, and won't be easily visible on PowerShell
        self.reporter.console.print(f"{indent}Command to be run: [magenta]{markup.escape(shlex.join(command))}[/magenta]", highlight=False)
        if input_file is not None or stdout_as_output:
            self.reporter.console.print(f"{indent}Command to be run including I/O redirection:")
            self.reporter.console.print(
                f"{indent}[magenta]{markup.escape(shlex.join(command))}[/magenta]"
                + (f" \\\n{indent_plus}[magenta]< {markup.escape(shlex.quote(str(input_file)))}[/magenta]" if input_file is not None else "")
                + (f" \\\n{indent_plus}[magenta]> {markup.escape(shlex.quote(str(output_file)))}[/magenta]" if stdout_as_output else ""),
                highlight=False,
            )

        return StdioCommandData(output_file, self.reporter.stdout, self.reporter.stderr)

    def maybe_launch_debugger(self, command: Sequence[str]) -> None:
        indent = indentation(self.reporter.depth)
        if self.reporter.debug_mode:
            self.reporter.console.print(f"{indent}This test did not complete successfully. We are going to launch the debugger now.")
            self.reporter.console.print(f"{indent}Debug command: [magenta]{markup.escape(shlex.join(command))}[/magenta]")
            launch_debugger = True
            while True:
                response = Prompt.ask(
                    prompt=f"{indent}Press [cyan]Enter[/cyan] to launch the debugger, or type 's' and then [cyan]Enter[/cyan] to skip",
                    console=self.reporter.console,
                )
                if response == "":
                    break
                if response[0].lower() == "s":
                    launch_debugger = False
                    break
            if launch_debugger:
                subprocess.run(command)

    def report_diff_result(self, test_identifier: str, goal_file: Path, output_file: Path, similarity: float, fuzzy: bool) -> None:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(
            f"{indent}Diff between [yellow]{markup.escape(shlex.quote(str(goal_file)))}[/yellow] and [yellow]{markup.escape(shlex.quote(str(output_file)))}[/yellow]"
        )
        # TODO Customize stdio directory here
        directory = TEST_DIRECTORY / "stdio/diffs"
        directory.mkdir(parents=True, exist_ok=True)
        html_output = directory / f"{test_identifier}.html"
        html_diff(goal_file, output_file, html_output)
        # NOTE: file:// will not be highlighted on Windows in PowerShell, Command Prompt, or Git Bash
        url = f"file://{html_output.resolve().absolute()}"
        self.reporter.console.print(f"{indent}Diff available to view at {url}")
        if self.reporter.debug_mode and similarity < 1.0:
            while True:
                response = Prompt.ask(
                    prompt=f"{indent}Press [cyan]Enter[/cyan] to continue, or type 'o' and then [cyan]Enter[/cyan] to open this diff view in the web browser",
                    console=self.reporter.console,
                )
                if response == "":
                    break
                if response[0].lower() == "o":
                    webbrowser.open(url)

    def report_file_exists_test(self, file: Path, expected: str) -> None:
        indent = indentation(self.reporter.depth)
        self.reporter.console.print(f"{indent}Checking for the existence of '{file}' and its containing type of data: {expected}")

    def report_file_exists_test_result(self, file: Path, expected: str, is_correct: bool, actual: Optional[str]) -> None:
        indent_plus = indentation(self.reporter.depth + 1)
        if is_correct:
            self.reporter.console.print(f"{indent_plus}[green]Actual type matches expected type![/green] Actual: {actual}")
        elif actual is not None:
            self.reporter.console.print(f"{indent_plus}[red]Actual type DOES NOT match expected type![/red] Actual: {actual}")
        else:
            self.reporter.console.print(f"{indent_plus}[red]File does not exist![/red]")


@dataclasses.dataclass
class ConsoleGraderReporter(GraderReporter):
    section: Section
    console: Console
    stdout: TextIO
    stderr: TextIO
    interactive: bool
    debug_mode: bool
    depth: int

    @property
    def test_reporter(self) -> TestReporter:
        return ConsoleTestReporter(self)

    def __report_test(self, max_points: int) -> None:
        test = self.section.node
        assert isinstance(test, Test)
        indent = indentation(self.depth)
        self.console.line()
        if self.depth == 0:
            self.console.print(
                (indent + header(self.depth, self.console.width))[: self.console.width],
                style="blue",
            )
        if max_points == 0:
            self.console.print(f"{arrow(self.depth)}[yellow]{markup.escape(self.section.name)}[/yellow] (ungraded)")
        else:
            self.console.print(f"{arrow(self.depth)}[yellow]{markup.escape(self.section.name)}[/yellow] is worth {max_points} points")
        self.console.print(
            (indent + header(self.depth, self.console.width))[: self.console.width],
            style="blue",
        )

    def __report_section(self, max_points: int) -> None:
        assert isinstance(self.section.node, list)
        sections: list[Section] = self.section.node
        assert len(sections) > 0
        assert isinstance(sections[0], Section)

        indent = indentation(self.depth)
        self.console.line()
        if self.depth == 0:
            self.console.print(
                (indent + header(self.depth, self.console.width))[: self.console.width],
                style="blue",
            )
        self.console.print(f"{arrow(self.depth)}[orange_red1]{self.section.name}[/orange_red1] is worth {max_points} points and has {len(sections)} parts")
        self.console.print(
            (indent + header(self.depth, self.console.width))[: self.console.width],
            style="blue",
        )

    def report_start(self, max_points: int) -> None:
        if isinstance(self.section.node, Test):
            self.__report_test(max_points)
        else:
            self.__report_section(max_points)

    def report_test_result(self, result: TestResult, points: int, max_points: int) -> None:
        indent = indentation(self.depth)
        if max_points == 0:
            self.console.print(f"{indent}Ungraded step completed successfully.")
        elif isinstance(result, FractionalTestResult):
            if result.total_count == 1:
                if result.success_count == 0:
                    self.console.print(
                        f"{indent}FAIL gives you {points} / {max_points} points",
                        style="red",
                    )
                else:
                    self.console.print(
                        f"{indent}SUCCESS gives you {points} / {max_points} points",
                        style="green",
                    )
            else:
                self.console.print(
                    f"{indent}Result is {result.success_count} / {result.total_count} which gives you {points} / {max_points} points",
                    style="green" if points > max_points // 2 else "red",
                )
        else:
            assert isinstance(result, PercentTestResult)
            self.console.print(f"Result is {result.percent:.2%} which gives you {points}")

    def subsection(self, section: Section) -> "GraderReporter":
        r = copy.copy(self)
        r.section = section
        r.depth += 1
        return r



import abc
import dataclasses
from typing import Union

from gradema.test import TestReporter


@dataclasses.dataclass
class FractionalTestResult:
    success_count: int
    total_count: int


@dataclasses.dataclass
class PercentTestResult:
    percent: float


TestResult = Union[FractionalTestResult, PercentTestResult]


class Test(abc.ABC):
    """
    This is the abstract class inherited by all tests.
    """

    @abc.abstractmethod
    def run(self, reporter: TestReporter) -> TestResult:
        pass



import subprocess

from gradema.test._reporter import TestReporter
from gradema.test._test import Test, TestResult, FractionalTestResult


class CommandTest(Test):
    def __init__(self, command: list[str], debug_command: list[str] | None = None):
        self.command = command
        self.debug_command = debug_command

    def run(self, reporter: TestReporter) -> TestResult:
        data = reporter.report_command(self.command)

        try:
            completed = subprocess.run(
                self.command,
                stdin=subprocess.DEVNULL,
                stdout=data.stdout,
                stderr=data.stderr,
            )
        except FileNotFoundError as e:
            reporter.log_unexpected_exception(e)
            reporter.log(f"Your environment may not be configured correctly. Could not find binary for {self.command[0]}")
            return FractionalTestResult(0, 1)

        # Note that these kinds of tests are susceptible to early-exit-attacks
        #   It is up to the user of this test to make sure that it cannot be exploited.
        if completed.returncode == 0:
            return FractionalTestResult(1, 1)
        if self.debug_command is not None:
            reporter.maybe_launch_debugger(self.debug_command)
        return FractionalTestResult(0, 1)



import io
from typing import TextIO


class IndentTextIOWrapper(io.BytesIO):
    """
    A wrapper around a TextIO object to put indentation before each line.

    This is a little messy because the subprocess package won't let you pass just any TextIO object to the run function,
    so you usually have to pass subprocess.PIPE, and then feed the piped output to this for it to actually do anything.

    TLDR; This would be cool to use, but it's hard to get to work.

    Additionally, putting correct type hints on this is difficult, so we need to figure that out

    Notes for later:
    MYPY_FORCE_COLOR=1
    """

    def __init__(self, stdout: TextIO, indent: str):
        super().__init__()
        self.stdout = stdout
        self.indent = indent
        self.__needs_indent = True

    def write(self, text):  # type: ignore
        for ascii_value in text:
            c = chr(ascii_value)
            if c == "\n":
                self.__needs_indent = True
            elif c != "\r" and self.__needs_indent:
                self.stdout.write(self.indent)
                self.__needs_indent = False
            self.stdout.write(c)

    def flush(self) -> None:
        self.stdout.flush()



from os import PathLike
from pathlib import Path

from . import TestReporter, Test, TestResult, FractionalTestResult


def create_file_exists_test(files: list[tuple[str | Path | PathLike[str], str]], mime: bool = False) -> Test:
    return FileExistsTest(files, mime)


class FileExistsTest(Test):

    def __init__(self, files: list[tuple[str | Path | PathLike[str], str]], mime: bool):
        self.files: list[tuple[str | Path | PathLike[str], str]] = files
        self.mime = mime

    def run(self, reporter: TestReporter) -> TestResult:
        success_count = 0
        total_count = 0
        for path_like, expected in self.files:
            path = path_like if isinstance(path_like, Path) else Path(path_like)
            reporter.report_file_exists_test(path, expected)
            total_count += 1
            output = None
            try:
                # noinspection PyUnresolvedReferences
                import magic

                # NOTE: magic.from_file(path) works when path is a pathlib.Path on Linux, but does not work on Windows.
                #   We must convert to string first!
                output = magic.from_file(str(path))
            except FileNotFoundError:
                pass
            except OSError as e:
                reporter.log_unexpected_exception(e)
            except ImportError as e:
                reporter.log_unexpected_exception(e)
                reporter.log("Unexpected import error")
            except AttributeError:
                # An AttributeError occurs on Windows when python-magic-bin is not installed
                #   `AttributeError: module 'magic' has no attribute 'from_file'
                # https://github.com/ahupp/python-magic?tab=readme-ov-file#installation
                reporter.log("libmagic is not installed. This test will fail. Make sure that python-magic-bin is installed if you are on Windows.")
            is_correct = output is not None and expected in output  # expected and output don't have to match, expected just has to be contained in the actual output
            if is_correct:
                success_count += 1
            reporter.report_file_exists_test_result(path, expected, is_correct, output)

        return FractionalTestResult(success_count, total_count)



import argparse
import sys
from pathlib import Path

from rich.console import Console
from rich.text import Text

from ._console import ConsoleGraderReporter
from gradema.section import Section
from gradema.grader import grade_section


RESULTS_PATH = Path("results.txt")


def save_score(points: int) -> None:
    # It is necessary to remove results.txt before writing to it because if a student wrote to results.txt and changed its permissions, we couldn't write to it
    RESULTS_PATH.unlink(missing_ok=True)
    with RESULTS_PATH.open("w") as file:
        print(points, file=file)


def welcome(console: Console) -> None:
    console.print("Welcome to Gradema!")
    console.print("This script will grandma you through completing this programming assignment")
    console.line()
    console.print(f"    [magenta]Magenta[/magenta] is bash/shell commands.")
    console.print(f"    [green]Green[/green] is used to indicate a pass.")
    console.print(f"    [red]Red[/red] is used to indicate a fail.")
    console.print(f"    [orange_red1]Orange[/orange_red1] marks the text of a header for a block of tests of a single type.")
    console.print(f"    [cyan]Cyan[/cyan] draws your attention to something you have to do..")
    console.line()
    console.print(Text.assemble("Actually ", ("read ALL the output generated by this script!", "cyan")))

    console.print(f"[blue]{'=' * console.width}[/blue]")


def run_grader(args: list[str], section: Section) -> int:
    if not section.is_pointed:
        raise ValueError("The root section must be pointed!")
    assert section.points is not None
    parser = argparse.ArgumentParser(
        prog="Gradema Autograder",
        epilog="Gradema - created by Lavender Shannon for Dr. Taylor",
    )
    parser.add_argument("-g", "--grade", action="store_true")
    parser.add_argument("-d", "--debug", action="store_true")
    try:
        parsed_args = parser.parse_args(args)
    except SystemExit as e:
        code = e.code
        assert isinstance(code, int)
        return code

    # Useful documentation here: https://rich.readthedocs.io/en/latest/markup.html
    console = Console(soft_wrap=True)
    welcome(console)

    interactive = True
    debug_mode = True
    if not parsed_args.debug:
        if parsed_args.grade:
            debug_mode = False
            # TODO determine if we are in an interactive terminal and set interactive accordingly
            # set interactive to False for now
            interactive = False
        else:
            try:
                console.print("Press [cyan]Enter[/cyan] now to continue")
                input()
            except EOFError:
                interactive = False
                debug_mode = False

        if interactive and debug_mode:  # if debug mode was set to False, then the --grade option must have been specified
            console.print(
                """Which of the following modes do you want to run in?
        1) debug+grade mode (d)
        2) grade only mode (g)
    Type 'd' or 'g', then hit [cyan]Enter[/cyan]."""
            )
            mode_input = input()
            debug_mode = mode_input == "" or mode_input[0].lower() != "g"

    reporter = ConsoleGraderReporter(
        section,
        console,
        sys.stdout,
        sys.stderr,
        interactive=interactive,
        debug_mode=debug_mode,
        depth=0,
    )
    points = grade_section(reporter, section.points, section)
    save_score(points)
    passing = points >= 70
    console.line()
    console.print("=" * console.width, style="green" if passing else "red")
    console.print(f"Your total grade is:\n{points}")
    console.line()
    if points == 100:
        console.print("It's perfect! Congratulations!", style="green")
        return 0
    elif passing:
        console.print("You're passing; decide how much you want the extra points!", style="green")
        return 0
    else:
        console.print("You're not passing yet.", style="red")
        return 1



from ._reporter import TestReporter, CommandData, StdioCommandData
from ._test import Test, FractionalTestResult, PercentTestResult, TestResult
from ._python import (
    create_python_test,
    create_python_pytest,
    create_python_format_check,
    create_python_format_check_from_path,
    create_python_type_check,
    create_mypy_command,
    create_python_stdio_test,
    create_python_traditional_stdio_test,
    create_python_traditional_arg_test,
)
from ._dummy import dummy_test
from ._exists import create_file_exists_test
from ._rust import RustProgram

__all__ = [
    "TestReporter",
    "CommandData",
    "StdioCommandData",
    "Test",
    "FractionalTestResult",
    "PercentTestResult",
    "TestResult",
    "create_python_test",
    "create_python_pytest",
    "create_python_stdio_test",
    "create_python_traditional_stdio_test",
    "create_python_format_check",
    "create_python_format_check_from_path",
    "create_python_type_check",
    "create_mypy_command",
    "dummy_test",
    "create_file_exists_test",
    "RustProgram",
    "create_python_traditional_arg_test",
]



import abc
import dataclasses
from pathlib import Path
from typing import TextIO, Optional, Sequence, Mapping

from gradema.test.argument import Argument, ResolvedArgumentInfo

"""
It's worth noting that whenever you see TextIO in this file, it likely doesn't really mean TextIO.
Since both sys.stdout and sys.stderr are TextIO, it makes sense to describe its type as such.
However, most implementations that reference fields from this file are actually using the subprocess module under the hood.
If you change the implementation here, you need to take special care to make sure the object you supply to callers works with the subprocess module.

https://docs.python.org/3/library/subprocess.html#popen-constructor
"""


@dataclasses.dataclass
class CommandData:
    stdout: TextIO
    """The stdout to pass to subprocess.run"""
    stderr: TextIO
    """The stderr to pass to subprocess.run"""


@dataclasses.dataclass
class StdioCommandData:
    """
    Represents what stdout and stderr should be set to on a command.
    stdout is represented by a file path because it is expected that the caller wants to be able to know how to read from that file after the command is executed.
    """

    stdout_file: Optional[Path]
    """The file that stdout should be directed to"""
    stdout: TextIO
    """The stdout to pass to subprocess.run assuming that stdout_file is None"""
    stderr: TextIO
    """The stderr to pass to subprocess.run"""


class TestReporter(abc.ABC):
    """
    A TestReporter is passed to all tests. This exposes a public API for tests to use and report what they are doing.
    A TestReporter's implementation may log what the test is doing to the console, or even perform an action depending on how the TestReporter is configured.
    """

    @abc.abstractmethod
    def log(self, message: str) -> None:
        """
        Log an informative message to the user.

        :param message: The message to log. No formatting is support
        """
        pass

    @abc.abstractmethod
    def log_unexpected_exception(self, exception: BaseException) -> None:
        """
        Log an unexpected exception

        :param exception: The exception to log
        """
        pass

    @abc.abstractmethod
    def resolve_command(self, command: Sequence[Argument], test_identifier: str) -> tuple[Sequence[str], Mapping[tuple[int, Argument], ResolvedArgumentInfo]]:
        """
        Resolves a list of Arguments by returning a list of strings.

        The returned value contains the resolved command and an argument info dictionary.
        The key of the argument info dictionary is a tuple of (index, Argument) where index is the index of the Argument in the command parameter.

        If you are wondering why ``command``'s type is  ``Sequence[Argument]`` rather than a ``list[Argument]``, it is because the Sequence type is covariant.
        That means that you can assign a ``Sequence[str]`` to a ``Sequence[Argument]``, which is necessary to make sure mypy type checking is happy,
        and correctly documents how the ``command`` argument is used: it is not modified. More information here: https://mypy.readthedocs.io/en/stable/common_issues.html#variance

        :param test_identifier: The identifier of the test being run
        :param command: The command to resolve
        :return: A tuple of (resolved command, argument info dictionary)
        :raises ValueError: A ValueError may be risen if the parameters supplied to this function do not give enough information to resolve a particular Argument.
                            This exception is currently not risen in any known implementation, but it's possible that optional parameters will be added in the future
                            that aid in resolving an Argument that is not created as of writing this docstring
        """
        pass

    @abc.abstractmethod
    def report_command(self, command: Sequence[str]) -> CommandData:
        """
        Reports that the given command WILL be run

        :param command: The command that will be run
        :return: A CommandData object containing parameters that determine where the stdout and stdin of the command should go.
        """
        pass

    @abc.abstractmethod
    def report_stdio_command(self, command: Sequence[str], test_identifier: str, input_file: Optional[Path], stdout_as_output: bool) -> StdioCommandData:
        """
        Reports that a stdio test command WILL be run.

        This is similar to report_command, but it is expected that the caller will run the command and direct its output to that specified by the returned object.

        :param command: The base command that is being run (not including any I/O redirection)
        :param test_identifier: An identifier for the test. This is not recommended to contain spaces or upper case characters
        :param input_file: The input file that will be directed into the command's stdin, or None if no specific input file is being used.
        :param stdout_as_output: True to use the command's stdout as an output, False to not
        :return: A StdioCommandData that contains the file to direct stdout to and the TextIO to direct stderr to.
        """
        pass

    @abc.abstractmethod
    def maybe_launch_debugger(self, command: Sequence[str]) -> None:
        """
        Calling this function will launch the debugger by running the given command only if we are in debugging mode.
        The implementation will likely prompt the user to make sure they want to launch the debugger

        :param command: The command to launch the debugger
        """
        pass

    @abc.abstractmethod
    def report_diff_result(self, test_identifier: str, goal_file: Path, output_file: Path, similarity: float, fuzzy: bool) -> None:
        """
        Reports the result of a diff and possibly opens a diff viewer depending on the configuration of this reporter.

        :param test_identifier: The identifier for the test
        :param goal_file: The goal file
        :param output_file: The output file that is supposed to contain the same contents of the goal file
        :param similarity: A number from 0 to 1 representing the similarity (1 is exactly the same). Non-fuzzy diffs should only pass 0 or 1.
        :param fuzzy: True if fuzzy, false otherwise
        """
        pass

    @abc.abstractmethod
    def report_file_exists_test(self, file: Path, expected: str) -> None:
        """
        Reports that a file existence test WILL BE performed.

        :param file:
        :param expected:
        :return:
        """
        pass

    @abc.abstractmethod
    def report_file_exists_test_result(self, file: Path, expected: str, is_correct: bool, actual: Optional[str]) -> None:
        """
        Reports that a file existence test has been performed and that the result is this.

        :param file: The file being tested for its existence
        :param expected: The expected type of the file (basically the output of `file <file>`
        :param is_correct: True if the actual matches the expected
        :param actual: The actual type of the file, or None if the file does not exist
        :return:
        """
        pass



__all__ = ["OUTPUT_FILE", "Argument", "ResolvedArgumentInfo", "OutputFile"]

from pathlib import Path
from typing import Union


class OutputFile:
    pass


OUTPUT_FILE = OutputFile()


Argument = str | OutputFile
ResolvedArgumentInfo = Union[None, Path]



import filecmp
import subprocess
from pathlib import Path
from typing import Optional, TextIO, Sequence

from gradema.test.argument import Argument, OUTPUT_FILE
from ._reporter import TestReporter
from ._test import Test, TestResult, FractionalTestResult


class StdioTest(Test):
    """
    A Standard Input/Output Test

    Traditionally, a stdio test would have no arguments passed to the command, it would have an input file, and a goal file.
    Output is piped to some output file that is not committed to version control.

    This new stdio test is very similar, but it can also do what a traditional arg test could do:
    take no input, and have custom arguments passed to the command.
    If you don't want anything directed into the command's stdin, then just assign None to input_file.
    If you want custom arguments passed to the command, give those custom arguments to the command argument (you may have to do that manually).
    """

    def __init__(self, command: Sequence[Argument], test_identifier: str, input_file: Optional[Path], goal_file: Path, stdout_as_output: bool):
        self.command = command
        self.test_identifier = test_identifier
        self.input_file = input_file
        self.goal_file = goal_file
        self.stdout_as_output = stdout_as_output

    def run(self, reporter: TestReporter) -> TestResult:
        command, argument_info_dictionary = reporter.resolve_command(self.command, self.test_identifier)
        data = reporter.report_stdio_command(command, self.test_identifier, self.input_file, self.stdout_as_output)
        output_file = data.stdout_file
        if output_file is None:  # output file will be None when stdout_as_output is False
            # By calling index() we are requiring that the command has an OUTPUT_FILE argument.
            #   If this is not the case, calling index() will raise a ValueError
            output_file_argument_index = self.command.index(OUTPUT_FILE)
            output_file_info = argument_info_dictionary[(output_file_argument_index, OUTPUT_FILE)]
            assert isinstance(output_file_info, Path), f"We expect OUTPUT_FILE's resolved argument info to be a Path! Instead it is {output_file_info}"
            output_file = output_file_info

        stdin: Optional[TextIO] = None
        try:
            if self.input_file is not None:
                stdin = self.input_file.open("r")
            stdout: Optional[TextIO] = None
            try:
                if data.stdout_file is not None:
                    stdout = data.stdout_file.open("w")
                try:
                    completed = subprocess.run(
                        command,
                        stdin=stdin if stdin is not None else subprocess.DEVNULL,
                        stdout=stdout if stdout is not None else data.stdout,
                        stderr=data.stderr,
                    )
                except FileNotFoundError as e:
                    reporter.log_unexpected_exception(e)
                    reporter.log(f"Your environment may not be configured correctly. Could not find binary for {command[0]}")
                    return FractionalTestResult(0, 1)
            finally:
                if stdout is not None:
                    stdout.close()
        finally:
            if stdin is not None:
                stdin.close()
        if completed.returncode != 0:
            reporter.log(f"Process had non-zero exit code: {completed.returncode}. Not going to execute a diff until the program can complete successfully.")
            return FractionalTestResult(0, 1)

        exactly_the_same = filecmp.cmp(self.goal_file, output_file)
        # TODO fuzzy diff stuff here
        reporter.report_diff_result(self.test_identifier, self.goal_file, output_file, 1.0 if exactly_the_same else 0.0, fuzzy=False)
        if exactly_the_same:
            return FractionalTestResult(1, 1)

        return FractionalTestResult(0, 1)



"""
The grader module contains utilities to grade a section and its subsections
"""

from ._reporter import GraderReporter
from ._grader import grade_section

__all__ = [
    "GraderReporter",
    "grade_section",
]



from os import PathLike
from pathlib import Path
from typing import Sequence, Optional

from gradema.test import Test
from gradema.test._command import CommandTest
from gradema.test._stdio import StdioTest
from gradema.test.argument import Argument


# noinspection PyMethodMayBeStatic
class RustProgram:
    def __init__(self, main_location: str | PathLike[str] | Path, binary_location: str | PathLike[str] | Path):
        self.main_location = main_location if isinstance(main_location, Path) else Path(main_location)
        self.binary_location = binary_location if isinstance(binary_location, Path) else Path(binary_location)

    def create_compile_step(self) -> Test:
        return CommandTest(["cargo", "build"])

    def create_stdio_test(self, test_identifier: str, args: Sequence[Argument], input_file: Optional[Path], goal_file: Path, stdout_as_output: bool) -> Test:
        command: list[Argument] = [str(self.binary_location)]
        command.extend(args)
        return StdioTest(command, test_identifier, input_file, goal_file, stdout_as_output)

    def create_traditional_stdio_test(self, test_identifier: str, input_file: Path, goal_file: Path) -> Test:
        return self.create_stdio_test(test_identifier, [], input_file, goal_file, stdout_as_output=True)

    def create_traditional_arg_test(self, test_identifier: str, args: Sequence[Argument], goal_file: Path) -> Test:
        return self.create_stdio_test(test_identifier, args, None, goal_file, stdout_as_output=False)

    def create_format_check(self) -> Test:
        """
        Creates a rustfmt check.

        Note that the resulting command will be something like ``rustfmt --check src/main.rs``.
        Even though this only includes the main file, all files referenced by main will also be format checked -- it's not necessary to list all the files here!
        https://github.com/rust-lang/rustfmt/issues/2426

        :return: A test that completes successfully when there are no formatting errors
        """
        return CommandTest(["rustfmt", "--check", str(self.main_location)])

    def create_cargo_command(self, arguments: list[str]) -> Test:
        return CommandTest(["cargo"] + arguments)



import dataclasses
from typing import Optional

from gradema.test import Test


SectionNode = list["Section"] | Test
"""Either a list of Sections, or a Test"""


@dataclasses.dataclass
class Section:
    """
    A section contains information about how many points it is worth.
    A section where points and weight are both none indicate an evenly weighted section.
    A section with points = 0 indicates an ungraded section.
    A section where points is non-zero indicates a pointed section.
    A section where weight is not None indicates a weighted section.
    """

    points: Optional[int]
    weight: Optional[float]
    name: str
    node: SectionNode

    @property
    def is_pointed(self) -> bool:
        return self.points is not None and self.points > 0

    @property
    def is_ungraded(self) -> bool:
        return self.points == 0

    @property
    def is_weighted(self) -> bool:
        return self.weight is not None

    @property
    def is_evenly_weighted(self) -> bool:
        return self.points is None and self.weight is None

    @classmethod
    def pointed(cls, points: int, name: str, node: SectionNode) -> "Section":
        if points <= 0:
            raise ValueError(f"points must be > 0. points: {points}")
        return Section(points, None, name, node)

    @classmethod
    def ungraded(cls, name: str, node: SectionNode) -> "Section":
        return Section(0, None, name, node)

    @classmethod
    def weighted(cls, weight: float, name: str, node: SectionNode) -> "Section":
        if weight <= 0:
            raise ValueError(f"weight must be > 0. weight: {weight}")
        return Section(None, weight, name, node)

    @classmethod
    def evenly_weighted(cls, name: str, node: SectionNode) -> "Section":
        return Section(None, None, name, node)
